"""Generate the SHACL schema based on the meta-model."""
import io
import textwrap
from typing import Tuple, Optional, List, Mapping, MutableMapping

from icontract import ensure, require

from aas_core_codegen import intermediate, specific_implementations, infer_for_schema
from aas_core_codegen.common import Stripped, Error, assert_never, Identifier
from aas_core_codegen.rdf_shacl import (
    naming as rdf_shacl_naming,
    common as rdf_shacl_common,
)
from aas_core_codegen.rdf_shacl.common import INDENT as I, INDENT2 as II, INDENT3 as III


@require(lambda prop, cls: id(prop) in cls.property_id_set)
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def _define_property_shape(
    prop: intermediate.Property,
    cls: intermediate.ClassUnion,
    url_prefix: Stripped,
    class_to_rdfs_range: rdf_shacl_common.ClassToRdfsRange,
    constraints_by_property: infer_for_schema.ConstraintsByProperty,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the shape of a property ``prop`` of the intermediate ``symbol``."""

    stmts = [Stripped("a sh:PropertyShape ;")]  # type: List[Stripped]

    # Resolve the type annotation to the actual value, regardless if the property is
    # mandatory or optional
    type_anno = intermediate.beneath_optional(prop.type_annotation)

    prop_name = rdf_shacl_naming.property_name(prop.name)

    rdfs_range = rdf_shacl_common.rdfs_range_for_type_annotation(
        type_annotation=type_anno, class_to_rdfs_range=class_to_rdfs_range
    )

    cls_name = rdf_shacl_naming.class_name(cls.name)

    stmts.append(Stripped(f"sh:path <{url_prefix}/{cls_name}/{prop_name}> ;"))

    if rdfs_range.startswith("rdf:") or rdfs_range.startswith("xsd:"):
        stmts.append(Stripped(f"sh:datatype {rdfs_range} ;"))
    elif rdfs_range.startswith("aas:"):
        stmts.append(Stripped(f"sh:class {rdfs_range} ;"))
    else:
        raise NotImplementedError(f"Unhandled namespace of the {rdfs_range=}")

    # region Define cardinality

    # noinspection PyUnusedLocal
    min_count = None  # type: Optional[int]
    max_count = None  # type: Optional[int]

    if isinstance(prop.type_annotation, intermediate.OptionalTypeAnnotation):
        if isinstance(type_anno, intermediate.ListTypeAnnotation):
            min_count = 0
            max_count = None

        elif isinstance(
            type_anno,
            (intermediate.OurTypeAnnotation, intermediate.PrimitiveTypeAnnotation),
        ):
            min_count = 0
            max_count = 1

        else:
            assert_never(type_anno)

    elif isinstance(prop.type_annotation, intermediate.ListTypeAnnotation):
        min_count = 0
        max_count = None

    elif isinstance(
        prop.type_annotation,
        (intermediate.OurTypeAnnotation, intermediate.PrimitiveTypeAnnotation),
    ):
        min_count = 1
        max_count = 1

    else:
        return None, Error(
            prop.parsed.node,
            f"(mristin, 2021-11-13): "
            f"We did not implement how to determine the cardinality based on the type "
            f"{prop.type_annotation}. If you see this message, it is time to implement "
            f"this logic.",
        )

    min_length = None  # type: Optional[int]
    max_length = None  # type: Optional[int]

    len_constraint = constraints_by_property.len_constraints_by_property.get(prop, None)

    if len_constraint is not None:
        if isinstance(type_anno, intermediate.ListTypeAnnotation):
            if len_constraint.min_value is not None:
                if len_constraint.min_value > 0 and isinstance(
                    prop.type_annotation, intermediate.OptionalTypeAnnotation
                ):
                    return None, Error(
                        prop.parsed.node,
                        f"(mristin, 2022-02-09): "
                        f"The property {prop.name} is optional, but the minCount "
                        f"is given. If you see this message, it is time to consider "
                        f"how to implement this logic; please contact the developers.",
                    )

                min_count = (
                    max(min_count, len_constraint.min_value)
                    if min_count is not None
                    else len_constraint.min_value
                )

            if len_constraint.max_value is not None:
                max_count = (
                    min(max_count, len_constraint.max_value)
                    if max_count is not None
                    else len_constraint.max_value
                )
        elif (
            isinstance(type_anno, intermediate.PrimitiveTypeAnnotation)
            and type_anno.a_type is intermediate.PrimitiveType.STR
        ):
            min_length = len_constraint.min_value
            max_length = len_constraint.max_value

        elif (
            isinstance(type_anno, intermediate.OurTypeAnnotation)
            and isinstance(type_anno.symbol, intermediate.ConstrainedPrimitive)
            and (
                type_anno.symbol.constrainee is intermediate.PrimitiveType.STR
                or type_anno.symbol.constrainee is intermediate.PrimitiveType.BYTEARRAY
            )
        ):
            min_length = len_constraint.min_value
            max_length = len_constraint.max_value
        else:
            return None, Error(
                prop.parsed.node,
                f"(mristin, 2022-02-09): "
                f"We did not implement how to specify the length constraint on the type "
                f"{type_anno}. If you see this message, it is time to implement "
                f"this logic.",
            )

    if min_count is not None:
        stmts.append(Stripped(f"sh:minCount {min_count} ;"))

    if max_count is not None:
        stmts.append(Stripped(f"sh:maxCount {max_count} ;"))

    if min_length is not None:
        stmts.append(Stripped(f"sh:minLength {min_length} ;"))

    if max_length is not None:
        stmts.append(Stripped(f"sh:maxLength {max_length} ;"))

    # endregion

    # region Define patterns

    pattern_constraints = constraints_by_property.patterns_by_property.get(prop, [])

    for pattern_constraint in pattern_constraints:
        pattern_literal = rdf_shacl_common.string_literal(pattern_constraint.pattern)

        stmts.append(Stripped(f"sh:pattern {pattern_literal} ;"))

    # endregion

    writer = io.StringIO()
    writer.write("sh:property [")
    for stmt in stmts:
        writer.write("\n")
        writer.write(textwrap.indent(stmt, I))

    writer.write("\n] ;")

    return Stripped(writer.getvalue()), None


# fmt: off
@require(
    lambda cls:
    not isinstance(cls, intermediate.Class)
    or not cls.is_implementation_specific
)
# fmt: on
def _define_for_class(
    cls: intermediate.ClassUnion,
    class_to_rdfs_range: rdf_shacl_common.ClassToRdfsRange,
    url_prefix: Stripped,
    constraints_by_property: infer_for_schema.ConstraintsByProperty,
) -> Tuple[Optional[Stripped], Optional[Error]]:
    """Generate the definition for the class ``cls``."""
    prop_blocks = []  # type: List[Stripped]
    errors = []  # type: List[Error]

    for prop in cls.properties:
        if prop.specified_for is not cls:
            continue

        prop_block, error = _define_property_shape(
            prop=prop,
            cls=cls,
            url_prefix=url_prefix,
            class_to_rdfs_range=class_to_rdfs_range,
            constraints_by_property=constraints_by_property,
        )

        if error is not None:
            errors.append(error)
        else:
            assert prop_block is not None
            prop_blocks.append(prop_block)

    if len(errors) > 0:
        return None, Error(
            None, f"Failed to generate the shape definition for {cls.name}", errors
        )

    shape_name = rdf_shacl_naming.class_name(Identifier(cls.name + "_shape"))
    cls_name = rdf_shacl_naming.class_name(cls.name)

    writer = io.StringIO()
    writer.write(
        f"""\
aas:{shape_name} a sh:NodeShape ;
{I}sh:targetClass aas:{cls_name} ;"""
    )

    for inheritance in cls.inheritances:
        subclass_shape_name = rdf_shacl_naming.class_name(
            Identifier(f"{inheritance.name}_shape")
        )

        writer.write(f"\n{I}rdfs:subClassOf aas:{subclass_shape_name} ;")

    if isinstance(cls, intermediate.AbstractClass):
        writer.write("\n")
        # pylint: disable=line-too-long
        writer.write(
            textwrap.indent(
                f'''\
sh:sparql [
{I}a sh:SPARQLConstraint ;
{I}sh:message "({shape_name}): An aas:{cls_name} is a abstract class. Please use one of the subclasses for the generation of instances."@en ;
{I}sh:prefixes aas: ;
{I}sh:select """
{II}SELECT ?this ?type
{II}WHERE {{
{III}?this rdf:type ?type .
{III}FILTER (?type = aas:{cls_name})
{II}}}
{I}""" ;
] ;''',
                I,
            )
        )
        # pylint: enable=line-too-long

    for block in prop_blocks:
        writer.write("\n")
        writer.write(textwrap.indent(block, I))

    writer.write("\n.")

    return Stripped(writer.getvalue()), None


@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
def generate(
    symbol_table: intermediate.SymbolTable,
    class_to_rdfs_range: rdf_shacl_common.ClassToRdfsRange,
    spec_impls: specific_implementations.SpecificImplementations,
    url_prefix: Stripped,
) -> Tuple[Optional[Stripped], Optional[List[Error]]]:
    """Generate the SHACL schema based on the ``symbol_table."""
    errors = []  # type: List[Error]

    preamble_key = specific_implementations.ImplementationKey("shacl/preamble.ttl")

    preamble = spec_impls.get(preamble_key, None)
    if preamble is None:
        errors.append(
            Error(
                None,
                f"The implementation snippet for the SHACL preamble "
                f"is missing: {preamble_key}",
            )
        )

    if len(errors) > 0:
        return None, errors

    assert preamble is not None
    blocks = [preamble]  # type: List[Stripped]

    constraints_by_class, some_errors = infer_for_schema.infer_constraints_by_class(
        symbol_table=symbol_table
    )

    if some_errors is not None:
        errors.extend(some_errors)

    if len(errors) > 0:
        return None, errors

    assert constraints_by_class is not None

    for symbol in sorted(
        symbol_table.symbols,
        key=lambda a_symbol: rdf_shacl_naming.class_name(a_symbol.name),
    ):
        # noinspection PyUnusedLocal
        block = None  # type: Optional[Stripped]

        if isinstance(symbol, intermediate.Enumeration):
            continue

        elif isinstance(symbol, intermediate.ConstrainedPrimitive):
            # NOTE (mristin, 2022-02-11):
            # We in-line the constraints from the constrained primitives directly in the
            # properties. We do not want to introduce separate entities for them as that
            # would unnecessarily pollute the ontology.
            continue

        elif isinstance(
            symbol, (intermediate.AbstractClass, intermediate.ConcreteClass)
        ):
            if symbol.is_implementation_specific:
                implementation_key = specific_implementations.ImplementationKey(
                    f"shacl/{symbol.name}/shape.ttl"
                )

                implementation = spec_impls.get(implementation_key, None)
                if implementation is None:
                    errors.append(
                        Error(
                            symbol.parsed.node,
                            f"The implementation snippet for "
                            f"the class {symbol.parsed.name} "
                            f"is missing: {implementation_key}",
                        )
                    )
                else:
                    blocks.append(implementation)

            else:
                block, error = _define_for_class(
                    cls=symbol,
                    class_to_rdfs_range=class_to_rdfs_range,
                    url_prefix=url_prefix,
                    constraints_by_property=constraints_by_class[symbol],
                )

                if error is not None:
                    errors.append(error)
                else:
                    assert block is not None
                    blocks.append(block)
        else:
            assert_never(symbol)

    if len(errors) > 0:
        return None, errors

    return Stripped("\n\n".join(blocks)), None



"""Generate C# code for de/serialization based on the intermediate representation."""

import io
import textwrap
import xml.sax.saxutils
from typing import Tuple, Optional, List

from icontract import ensure

from aas_core_codegen import intermediate
from aas_core_codegen.common import Error, Stripped, Identifier
from aas_core_codegen.csharp import common as csharp_common, naming as csharp_naming
from aas_core_codegen.csharp.common import INDENT as I, INDENT2 as II, INDENT3 as III


def _generate_enum_to_and_from_string(
    enumeration: intermediate.Enumeration,
) -> Stripped:
    """Generate the methods for de/serializing enumeration from/to a string."""
    blocks = []  # type: List[Stripped]

    name = csharp_naming.enum_name(enumeration.name)

    # region To-string-map

    to_str_map_name = csharp_naming.private_property_name(
        Identifier(f"{enumeration.name}_to_string")
    )

    to_str_map_writer = io.StringIO()
    to_str_map_writer.write(
        f"private static readonly Dictionary<Aas.{name}, string> {to_str_map_name} = (\n"
        f"{I}new Dictionary<Aas.{name}, string>()\n"
        f"{I}{{\n"
    )

    for i, literal in enumerate(enumeration.literals):
        literal_name = csharp_naming.enum_literal_name(literal.name)
        to_str_map_writer.write(
            f"{II}{{ Aas.{name}.{literal_name}, "
            f"{csharp_common.string_literal(literal.value)} }}"
        )

        if i < len(enumeration.literals) - 1:
            to_str_map_writer.write(",")

        to_str_map_writer.write("\n")

    to_str_map_writer.write(f"{I}}});")

    blocks.append(Stripped(to_str_map_writer.getvalue()))

    # endregion

    # region To-string-method

    to_str_name = csharp_naming.method_name(Identifier("to_string"))

    to_str_writer = io.StringIO()
    to_str_writer.write(
        f"""\
/// <summary>
/// Retrieve the string representation of <paramref name="that" />.
/// </summary>
/// <remarks>
/// If <paramref name="that" /> is not a valid literal, return <c>null</c>.
/// </remarks>
public static string? {to_str_name}(Aas.{name}? that)
{{
{I}if (!that.HasValue)
{I}{{
{II}return null;
{I}}}
{I}else
{I}{{
{II}if ({to_str_map_name}.TryGetValue(that.Value, out string? value))
{II}{{
{III}return value;
{II}}}
{II}else
{II}{{
{III}return null;
{II}}}
{I}}}
}}"""
    )

    blocks.append(Stripped(to_str_writer.getvalue()))

    # endregion

    # region From-string-map

    from_str_map_name = csharp_naming.private_property_name(
        Identifier(f"{enumeration.name}_from_string")
    )

    from_str_map_writer = io.StringIO()
    from_str_map_writer.write(
        f"private static readonly Dictionary<string, Aas.{name}> {from_str_map_name} = (\n"
        f"{I}new Dictionary<string, Aas.{name}>()\n"
        f"{I}{{\n"
    )

    for i, literal in enumerate(enumeration.literals):
        literal_name = csharp_naming.enum_literal_name(literal.name)
        from_str_map_writer.write(
            f"{II}{{ {csharp_common.string_literal(literal.value)}, "
            f"Aas.{name}.{literal_name} }}"
        )

        if i < len(enumeration.literals) - 1:
            from_str_map_writer.write(",")

        from_str_map_writer.write("\n")

    from_str_map_writer.write(f"{I}}});")

    blocks.append(Stripped(from_str_map_writer.getvalue()))

    # endregion

    # region From-string-method

    from_str_name = csharp_naming.method_name(
        Identifier(f"{enumeration.name}_from_string")
    )

    from_str_writer = io.StringIO()
    from_str_writer.write(
        f"""\
/// <summary>
/// Parse the string representation of <see cref={xml.sax.saxutils.quoteattr(name)} />.
/// </summary>
/// <remarks>
/// If <paramref name="text" /> is not a valid string representation
/// of a literal of <see cref={xml.sax.saxutils.quoteattr(name)} />,
/// return <c>null</c>.
/// </remarks>
public static Aas.{name}? {from_str_name}(string text)
{{
{I}if ({from_str_map_name}.TryGetValue(text, out {name} value))
{I}{{
{II}return value;
{I}}}
{I}else
{I}{{
{II}return null;
{I}}}
}}"""
    )

    blocks.append(Stripped(from_str_writer.getvalue()))

    # endregion

    return Stripped("\n\n".join(blocks))


# fmt: off
@ensure(lambda result: (result[0] is not None) ^ (result[1] is not None))
@ensure(
    lambda result:
    not (result[0] is not None) or result[0].endswith('\n'),
    "Trailing newline mandatory for valid end-of-files"
)
# fmt: on
def generate(
    symbol_table: intermediate.SymbolTable, namespace: csharp_common.NamespaceIdentifier
) -> Tuple[Optional[str], Optional[List[Error]]]:
    """
    Generate the C# code for the general serialization.

    The ``namespace`` defines the AAS C# namespace.
    """
    blocks = [
        csharp_common.WARNING,
        Stripped(
            f"""\
using System.Collections.Generic;  // can't alias

using Aas = {namespace};"""
        ),
    ]

    stringification_blocks = []  # type: List[Stripped]

    for symbol in symbol_table.symbols:
        if not isinstance(symbol, intermediate.Enumeration):
            continue

        stringification_blocks.append(
            _generate_enum_to_and_from_string(enumeration=symbol)
        )

    writer = io.StringIO()
    writer.write(
        f"""\
namespace {namespace}
{{
{I}public static class Stringification
{I}{{
"""
    )

    for i, stringification_block in enumerate(stringification_blocks):
        if i > 0:
            writer.write("\n\n")

        writer.write(textwrap.indent(stringification_block, II))

    writer.write(f"\n{I}}}  // public static class Stringification")
    writer.write(f"\n}}  // namespace {namespace}")

    blocks.append(Stripped(writer.getvalue()))

    blocks.append(csharp_common.WARNING)

    out = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            out.write("\n\n")

        assert not block.startswith("\n")
        assert not block.endswith("\n")
        out.write(block)

    out.write("\n")

    return out.getvalue(), None



"""Stringify the parsed meta-model."""
from typing import Optional, Union

from aas_core_codegen import stringify
from aas_core_codegen.parse import _types
from aas_core_codegen.parse import tree
from aas_core_codegen.parse._types import (
    AbstractClass,
    Argument,
    AtomicTypeAnnotation,
    ConcreteClass,
    Contract,
    Contracts,
    Default,
    Enumeration,
    EnumerationLiteral,
    Property,
    SelfTypeAnnotation,
    Snapshot,
    SubscriptedTypeAnnotation,
    Symbol,
    SymbolTable,
    TypeAnnotation,
    UnverifiedSymbolTable,
    Description,
    Invariant,
    ImplementationSpecificMethod,
    UnderstoodMethod,
    ConstructorToBeUnderstood,
    Serialization,
    MetaModel,
    ReferenceInTheBook,
)


def _stringify_atomic_type_annotation(
    that: AtomicTypeAnnotation,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("identifier", that.identifier),
            stringify.PropertyEllipsis("node", that.node),
        ],
    )

    return result


def _stringify_self_type_annotation(
    that: SelfTypeAnnotation,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[],
    )

    return result


def _stringify_subscripted_type_annotation(
    that: SubscriptedTypeAnnotation,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("identifier", that.identifier),
            stringify.Property("subscripts", list(map(_stringify, that.subscripts))),
            stringify.PropertyEllipsis("node", that.node),
        ],
    )

    return result


def _stringify_description(that: Description) -> stringify.Entity:
    return stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.PropertyEllipsis("document", that.document),
            stringify.PropertyEllipsis("node", that.node),
        ],
    )


def _stringify_property(that: Property) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property("type_annotation", _stringify(that.type_annotation)),
            stringify.Property("description", _stringify(that.description)),
            stringify.PropertyEllipsis("node", that.node),
        ],
    )

    return result


def _stringify_default(that: Default) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[stringify.PropertyEllipsis("node", that.node)],
    )

    return result


def _stringify_argument(that: Argument) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property("type_annotation", _stringify(that.type_annotation)),
            stringify.Property(
                "default",
                _stringify(that.default),
            ),
            stringify.PropertyEllipsis("node", that.node),
        ],
    )

    return result


def _stringify_invariant(that: Invariant) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("description", that.description),
            stringify.Property("body", tree.dump(that.body)),
            stringify.PropertyEllipsis("node", that.node),
        ],
    )

    return result


def _stringify_contract(that: Contract) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("args", that.args),
            stringify.Property("description", that.description),
            stringify.PropertyEllipsis("body", tree.dump(that.body)),
            stringify.PropertyEllipsis("node", that.node),
        ],
    )

    return result


def _stringify_snapshot(that: Snapshot) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("args", that.args),
            stringify.Property("name", that.name),
            stringify.Property("body", tree.dump(that.body)),
            stringify.PropertyEllipsis("node", that.node),
        ],
    )

    return result


def _stringify_contracts(that: Contracts) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property(
                "preconditions", list(map(_stringify, that.preconditions))
            ),
            stringify.Property("snapshots", list(map(_stringify, that.snapshots))),
            stringify.Property(
                "postconditions",
                list(map(_stringify, that.postconditions)),
            ),
        ],
    )

    return result


def _stringify_implementation_specific_method(
    that: ImplementationSpecificMethod,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property("verification", that.verification),
            stringify.Property("arguments", list(map(_stringify, that.arguments))),
            stringify.Property("returns", _stringify(that.returns)),
            stringify.Property("description", _stringify(that.description)),
            stringify.Property("contracts", _stringify(that.contracts)),
            stringify.PropertyEllipsis("node", that.node),
            stringify.PropertyEllipsis("arguments_by_name", that.arguments_by_name),
        ],
    )

    return result


def _stringify_understood_method(
    that: UnderstoodMethod,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property("verification", that.verification),
            stringify.Property("arguments", list(map(_stringify, that.arguments))),
            stringify.Property("returns", _stringify(that.returns)),
            stringify.Property("description", _stringify(that.description)),
            stringify.Property("contracts", _stringify(that.contracts)),
            stringify.Property("body", list(map(tree.dump, that.body))),
            stringify.PropertyEllipsis("node", that.node),
            stringify.PropertyEllipsis("arguments_by_name", that.arguments_by_name),
        ],
    )

    return result


def _stringify_constructor_to_be_understood(
    that: ConstructorToBeUnderstood,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property("verification", that.verification),
            stringify.Property("arguments", list(map(_stringify, that.arguments))),
            stringify.Property("returns", _stringify(that.returns)),
            stringify.Property("description", _stringify(that.description)),
            stringify.Property("contracts", _stringify(that.contracts)),
            stringify.PropertyEllipsis("node", that.node),
            stringify.PropertyEllipsis("arguments_by_name", that.arguments_by_name),
            stringify.PropertyEllipsis("body", that.body),
        ],
    )

    return result


def _stringify_serialization(
    that: Serialization,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("with_model_type", that.with_model_type),
        ],
    )

    return result


def _stringify_reference_in_the_book(
    that: ReferenceInTheBook,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("section", list(that.section)),
            stringify.Property("index", that.index),
            stringify.Property("fragment", that.fragment),
        ],
    )

    return result


def _stringify_abstract_class(that: AbstractClass) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property(
                "is_implementation_specific", that.is_implementation_specific
            ),
            stringify.Property("inheritances", that.inheritances),
            stringify.Property("properties", list(map(_stringify, that.properties))),
            stringify.Property("methods", list(map(_stringify, that.methods))),
            stringify.Property("invariants", list(map(_stringify, that.invariants))),
            stringify.Property("serialization", _stringify(that.serialization)),
            stringify.Property(
                "reference_in_the_book", _stringify(that.reference_in_the_book)
            ),
            stringify.Property("description", _stringify(that.description)),
            stringify.PropertyEllipsis("node", that.node),
            stringify.PropertyEllipsis("properties_by_name", that.properties_by_name),
            stringify.PropertyEllipsis("methods_by_name", that.methods_by_name),
        ],
    )

    return result


def _stringify_concrete_class(that: ConcreteClass) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property(
                "is_implementation_specific", that.is_implementation_specific
            ),
            stringify.Property("inheritances", that.inheritances),
            stringify.Property("properties", list(map(_stringify, that.properties))),
            stringify.Property("methods", list(map(_stringify, that.methods))),
            stringify.Property("invariants", list(map(_stringify, that.invariants))),
            stringify.Property("serialization", _stringify(that.serialization)),
            stringify.Property(
                "reference_in_the_book", _stringify(that.reference_in_the_book)
            ),
            stringify.Property("description", _stringify(that.description)),
            stringify.PropertyEllipsis("node", that.node),
            stringify.PropertyEllipsis("properties_by_name", that.properties_by_name),
            stringify.PropertyEllipsis("methods_by_name", that.methods_by_name),
        ],
    )

    return result


def _stringify_enumeration_literal(
    that: EnumerationLiteral,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property("value", that.value),
            stringify.Property("description", _stringify(that.description)),
            stringify.PropertyEllipsis("node", that.node),
        ],
    )

    return result


def _stringify_enumeration(that: Enumeration) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("name", that.name),
            stringify.Property("is_superset_of", that.is_superset_of),
            stringify.Property("literals", list(map(_stringify, that.literals))),
            stringify.Property(
                "reference_in_the_book", _stringify(that.reference_in_the_book)
            ),
            stringify.Property("description", _stringify(that.description)),
            stringify.PropertyEllipsis("node", that.node),
            stringify.PropertyEllipsis("literals_by_name", that.literals_by_name),
        ],
    )

    return result


def _stringify_meta_model(
    that: MetaModel,
) -> stringify.Entity:
    result = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("description", _stringify(that.description)),
            stringify.Property("book_url", that.book_url),
            stringify.Property("book_version", that.book_version),
        ],
    )

    return result


def _stringify_unverified_symbol_table(
    that: UnverifiedSymbolTable,
) -> stringify.Entity:
    entity = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("symbols", list(map(_stringify, that.symbols))),
            stringify.Property(
                "verification_functions",
                list(map(_stringify, that.verification_functions)),
            ),
            stringify.Property("meta_model", _stringify(that.meta_model)),
        ],
    )

    return entity


def _stringify_symbol_table(that: SymbolTable) -> stringify.Entity:
    entity = stringify.Entity(
        name=that.__class__.__name__,
        properties=[
            stringify.Property("symbols", list(map(_stringify, that.symbols))),
            stringify.Property(
                "verification_functions",
                list(map(_stringify, that.verification_functions)),
            ),
            stringify.Property("meta_model", _stringify(that.meta_model)),
        ],
    )

    return entity


Dumpable = Union[
    Argument,
    AtomicTypeAnnotation,
    ConcreteClass,
    ConstructorToBeUnderstood,
    Contract,
    Contracts,
    Default,
    Description,
    Enumeration,
    EnumerationLiteral,
    ImplementationSpecificMethod,
    Invariant,
    MetaModel,
    Property,
    ReferenceInTheBook,
    SelfTypeAnnotation,
    Serialization,
    Snapshot,
    SubscriptedTypeAnnotation,
    Symbol,
    SymbolTable,
    TypeAnnotation,
    UnderstoodMethod,
    UnverifiedSymbolTable,
]

stringify.assert_all_public_types_listed_as_dumpables(
    dumpable=Dumpable, types_module=_types
)

_DISPATCH = {
    AbstractClass: _stringify_abstract_class,
    Argument: _stringify_argument,
    AtomicTypeAnnotation: _stringify_atomic_type_annotation,
    ConcreteClass: _stringify_concrete_class,
    ConstructorToBeUnderstood: _stringify_constructor_to_be_understood,
    Contract: _stringify_contract,
    Contracts: _stringify_contracts,
    Default: _stringify_default,
    Description: _stringify_description,
    Enumeration: _stringify_enumeration,
    EnumerationLiteral: _stringify_enumeration_literal,
    ImplementationSpecificMethod: _stringify_implementation_specific_method,
    Invariant: _stringify_invariant,
    MetaModel: _stringify_meta_model,
    Property: _stringify_property,
    ReferenceInTheBook: _stringify_reference_in_the_book,
    SelfTypeAnnotation: _stringify_self_type_annotation,
    Serialization: _stringify_serialization,
    Snapshot: _stringify_snapshot,
    SubscriptedTypeAnnotation: _stringify_subscripted_type_annotation,
    SymbolTable: _stringify_symbol_table,
    UnderstoodMethod: _stringify_understood_method,
    UnverifiedSymbolTable: _stringify_unverified_symbol_table,
}

stringify.assert_dispatch_exhaustive(dispatch=_DISPATCH, dumpable=Dumpable)


def _stringify(that: Optional[Dumpable]) -> Optional[stringify.Entity]:
    """Dispatch to the correct ``_stringify_*`` method."""
    if that is None:
        return None

    stringify_func = _DISPATCH.get(that.__class__, None)
    if stringify_func is None:
        raise AssertionError(
            f"No stringify function could be found for the class {that.__class__}"
        )

    stringified = stringify_func(that)  # type: ignore
    assert isinstance(stringified, stringify.Entity)
    stringify.assert_compares_against_dict(stringified, that)

    return stringified


def dump(that: Optional[Dumpable]) -> str:
    """Produce a string representation of the ``dumpable`` for testing or debugging."""
    if that is None:
        return repr(None)

    stringified = _stringify(that)
    return stringify.dump(stringified)
